/* Copyright (c) 2012 Sven Herrmann and Andreas Spillner

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_MAXPLUSRATIONAL_H
#define _POLYMAKE_MAXPLUSRATIONAL_H

#include "polymake/Rational.h"

namespace pm {

class MaxPlusRational {
protected:
  Rational value;
  
public:
  // Initializes to neutral element which is -inf.
   MaxPlusRational() : value(std::numeric_limits<Rational>::min()){}
    
   MaxPlusRational(const int a) {
      if (a==0) value=std::numeric_limits<Rational>::min();
      else if (a==1) value=0;
      else value=a;
   }
   
   MaxPlusRational(const Rational& a):value(a) {}
      
   MaxPlusRational(const Integer& num, const Integer& den):value(num,den) {}
   
   MaxPlusRational(long num, long den):value(num,den){}
   
   MaxPlusRational(const Integer& num, long den):value(num,den){}
   
   MaxPlusRational(long num, const Integer& den):value(num,den){}
   
   MaxPlusRational(const double& b):value(b){}
   
   explicit MaxPlusRational(const char* s):value(s){}
   
   
   friend int isfinite(const MaxPlusRational& a)
   {
      return isfinite(a.value);
   }
   
   friend int isinf(const MaxPlusRational& a)
   {
     return isinf(a.value);
   }

   void set(const Integer& num, const Integer& den)
   {
     value.set(num,den);
   }

   void set(long num, long den)
   {
     value.set(num,den);
   }

protected:
   void _set(const char *s);
public:

   /** Conversion from a printable representation.
       Numerator and denominator are expected delimited by `/'.
       Omitted denominator assumed equal to 1.
   */
   MaxPlusRational& set(const char *s)
   {
     value.set(s);
     return *this;
   }

   MaxPlusRational& operator= (const MaxPlusRational& b)
   {
     value=b.value;
     return *this;
   }

   /// Swaps the values.
   void swap(MaxPlusRational& b)
   {
     value.swap(b.value);  
   }


   // Comparison with -inf.
   bool operator!() const { return isinf(value)!=-1; }

   // Comparison with -inf.
   operator bool() const { return isinf(value)==-1; }

   MaxPlusRational& operator+= (const MaxPlusRational& b)
   {
     if (b.value>value) value=b.value;
     return *this;
   }

   MaxPlusRational& operator*= (const MaxPlusRational& b)
   {
     value+=b.value;
     return *this;
   }

   MaxPlusRational& operator/= (const MaxPlusRational& b)
   {
     value-=b;
     return *this;
   }

   friend MaxPlusRational operator+ (const MaxPlusRational& a, const MaxPlusRational& b)
   {
     if(a.value>b.value) return a;
     return b;
   }

   friend MaxPlusRational operator* (const MaxPlusRational& a, const MaxPlusRational& b)
   {
      return MaxPlusRational(a.value+b.value);
   }

   friend MaxPlusRational operator/ (const MaxPlusRational& a, const MaxPlusRational& b)
   {
     return MaxPlusRational(a.value-b.value);
   }


   // Comparison. The magnitude of the return value is arbitrary, only its sign is relevant.
   int compare(const MaxPlusRational& b) const
   {
     return value.compare(b.value);
   }

   friend bool operator== (const MaxPlusRational& a, const MaxPlusRational& b)
   {
     return a.value==b.value;
   }

   
   Rational get_value() const { return value; }

   // FIXME: write "infinity".  For the linear algebra methods it is strictly necessary
   // to have *this* interpretion of ZERO and ONE, regardless how unintuitive it may look like. 
   static MaxPlusRational ZERO() {return MaxPlusRational(0); }

   static MaxPlusRational ONE() {return MaxPlusRational(1); }


   template <typename> friend class spec_object_traits;

};

inline bool operator!= (const MaxPlusRational& a, const MaxPlusRational& b) { return !(a==b); }
inline bool operator< (const MaxPlusRational& a, const MaxPlusRational& b) { return a.compare(b)<0; }
inline bool operator> (const MaxPlusRational& a, const MaxPlusRational& b) { return a.compare(b)>0; }
inline bool operator<= (const MaxPlusRational& a, const MaxPlusRational& b) { return a.compare(b)<=0; }
inline bool operator>= (const MaxPlusRational& a, const MaxPlusRational& b) { return a.compare(b)>=0; }



template <>
struct spec_object_traits< MaxPlusRational >
  : spec_object_traits<is_scalar> {
  typedef Rational serialized;

  static serialized& serialize(MaxPlusRational& a)
  {
    return a.value;
  }

  static const serialized& serialize(const MaxPlusRational& a)
  {
    return a.value;
  }

  static const MaxPlusRational& zero()
  {
    return MaxPlusRational::ZERO();
  }

  static const MaxPlusRational& one()
  {
    return MaxPlusRational::ONE();
  }
};

}//end namespace pm



namespace std {

   inline void swap(pm::MaxPlusRational& r1, pm::MaxPlusRational& r2) { r1.swap(r2); }

} // end namespace std

namespace polymake {
   using pm::MaxPlusRational;
}

#endif // _POLYMAKE_MAXPLUSRATIONAL_H

// Local Variables:
// mode:C++
// c-basic-offset:3
// End:
