/* Copyright (c) 2012 Sven Herrmann and Andreas Spillner

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   */

#include "polymake/client.h"
#include "polymake/Array.h"
#include "polymake/Graph.h"
#include "polymake/Set.h"
#include "polymake/RandomGenerators.h"
#include "polymake/RandomSubset.h"



namespace polymake { namespace graph {

namespace {

void next_split(const int start, const int end, UniformlyRandom<double> & random_source, Graph<> &G, const int current_node,Entire<Array<Set<int> > >::iterator &splits) {
   const int n=end-start-1;
   if (n<0) return;
   const int middle=start+n*random_source.get();//the last element assigned to the left set
   const int left_node=G.add_node();
   const int right_node=G.add_node();
   G.edge(left_node,current_node);
   G.edge(right_node,current_node);
   *(++splits)=range(start,middle);
   *(++splits)=range(middle+1,end);
//   cerr<<range(start,middle)<<endl;
//   cerr<<range(middle+1,end)<<endl;
   next_split(start,middle,random_source,G,left_node,splits);
   next_split(middle+1,end,random_source,G,right_node,splits);
}

} 

std::pair<perl::Object, Array<Set<int> > > rand_tree_pair(const int n, perl::OptionSet options)
{
   RandomSeed seed(options["seed"]);
   UniformlyRandom<double> random_source(seed);
   Graph<> G;
   Array<Set<int> > splits(2*n-3);
   Entire<Array<Set<int> > >::iterator s=entire(splits);
   const int first_node=G.add_node();
   const int middle=(n-2)*random_source.get();//the last element assigned to the left set
   const int left_node=G.add_node();
   const int right_node=G.add_node();
   G.edge(left_node,first_node);
   G.edge(right_node,first_node);
   *s=range(0,middle); 
//   cerr<<range(0,middle)<<endl;
   next_split(0,middle,random_source,G,left_node,s);
   next_split(middle+1,n-1,random_source,G,right_node,s);
//   cerr<<"start"<<endl;
  
   perl::Object g("Graph");
   g.take("ADJACENCY")<<G; 
   g.take("CONNECTED")<<true;
   g.take("CONNECTIVITY")<<1; 
   g.take("TRIANGLE_FREE")<<true;
   g.take("BIPARTITE")<<true;
//   cerr<<splits<<endl;
   return std::pair<perl::Object, Array<Set<int> > >(g,splits);
}

perl::Object rand_tree(const int n, perl::OptionSet options)
{
return rand_tree_pair(n,options).first;
}


Array<Set<int> > rand_tree_split_system(const int n, perl::OptionSet options)
{
   RandomSeed seed(options["seed"]);
   RandomPermutation<> random_permutation(n,seed);
   Array<int> perm(n);
   Entire<RandomPermutation<> >::const_iterator rp=entire(random_permutation);
   for (int i=0; i<n;++i,++rp) perm[i]=*rp;
   const Array<Set<int> > splits(rand_tree_pair(n,options).second);
   const int n_splits=splits.size();
   Array<Set<int> > permuted_splits(n_splits);
   for (int i=0; i<n_splits;++i) {
     Set<int> s;
     for (Entire<Set<int> >::const_iterator j=entire(splits[i]);!j.at_end();++j) 
       s.insert(perm[*j]);
     permuted_splits[i]=s;
  }
  return permuted_splits;
}




UserFunction4perl("# @category Producing from scratch"
                  "# Constructs a random binary tree with //n// leaves."
                  "# @param Int n"
                  "# @option Int seed controls the outcome of the random number generator;"
                  "#   fixing a seed number guarantees the same outcome. "
                  "# @return Graph",
                  &rand_tree, "rand_tree($;{seed => undef})");


UserFunction4perl("# @category Producing from scratch"
                  "# Constructs a split system defining random binary tree with //n// (labeled) leaves."
                  "# @param Int n"
                  "# @option Int seed controls the outcome of the random number generator;"
                  "#   fixing a seed number guarantees the same outcome. "
                  "# @return Array<Set<Int>>",
                  &rand_tree_split_system, "rand_tree_split_system($;{seed => undef})");



} }

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
