/* Copyright (c) 2012 Sven Herrmann and Andreas Spillner

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "polymake/client.h"
#include "polymake/hash_map"
#include "polymake/SparseMatrix.h"
#include "polymake/Graph.h"
#include <map>
#include "polymake/Matrix.h"
#include "polymake/Set.h"
#include "polymake/linalg.h"

namespace polymake { namespace metric {

namespace {

template <typename Scalar>
bool find_way(const int current_node,const Scalar& current_improv,const Graph<> &G,const NodeMap<Undirected,Vector<Scalar> >& coordinates,const int j,int& best_node,Scalar& best_improv)
{
   //cerr<<"now at "<<current_node<<" with "<<current_improv<<endl;
   bool found=false;
   for (Entire< Graph<>::adjacent_node_list >::const_iterator neighbor=entire(G.adjacent_nodes(current_node));!neighbor.at_end();++neighbor) {
     if (*neighbor==j) return true;
     //cerr<<coordinates[current_node]<<endl<<coordinates[*neighbor]<<endl;
     const Scalar improv =coordinates[current_node][j]-coordinates[*neighbor][j];
     if (improv>0) {
       //cerr<<"neighbor "<<(*neighbor)<<" with "<<improv<<endl;
       if (find_way(*neighbor,current_improv+improv,G,coordinates,j,best_node,best_improv)) return true;
       found=true;
     }
   }
   if (!found && current_improv>best_improv) {
        best_improv=current_improv;
        best_node=current_node;
        //cerr<<"new best node: "<<current_node<<" ("<<coordinates[current_node]<<") with improv "<<best_improv<<endl;
     }
   return false;
}

} //unnamed namespace


template <typename Scalar>
perl::Object find_realization(const Matrix<Scalar>& dist)
{
  const int d(dist.cols());
  const int n_facets=d*(d+1)/2;
  
  std::multimap<Scalar,std::pair<int,int> > distance_ordering;


  SparseMatrix<Scalar> A(n_facets,d);
  Vector<Scalar> b(n_facets);
  int index=d;
  for (int i=0; i<d; ++i) {
    A(i,i)=-1;
    for (int j=i+1; j<d; ++j) {
      A(index,i)=A(index,j)=-1;
      b[index++]=-dist(i,j);
      distance_ordering.insert(std::pair<Scalar,std::pair<int,int> >(dist(i,j),std::pair<int,int>(i,j)));
    }
  }
  //Ax<=b now defines the tight span

  // the Graph realising dist
  Graph<Undirected> G(d);
  EdgeMap<Undirected,Scalar> edge_length(G);
  Array<int> elements(d);

  NodeMap<Undirected,Vector<Scalar> > coordinates(G);
  NodeMap<Undirected, Set<int> > one_basis(G);
  hash_map<Vector<Scalar>, int> v_map;  
   

  //the default vertices
  for (int i=0; i<d; ++i){
    elements[i]=i;
    coordinates[i]=dist.row(i);
    v_map[dist.row(i)]=i;
    //a basis for the vertex
    Set<int> start_B=scalar2set(i);
    for (int j=0; j<i; ++j) start_B.insert(d+j*(2*d-j-1)/2+i-j-1);
    for (int j=i+1; j<d; ++j) start_B.insert(d+i*(2*d-i-1)/2+j-i-1);
    one_basis[i]=start_B;
  }
 
  int current_node;
  Set<int> B;
  Vector<Scalar> costs(d);

  for (typename std::multimap<Scalar,std::pair<int,int> >::const_iterator iti=distance_ordering.begin(); iti!=distance_ordering.end(); ++iti) {
      int i=iti->second.first;
      int j=iti->second.second;
      current_node=i;
      B=one_basis[i];
      //cerr<<endl<<"find way from "<<i<<" to "<<j<<":"<<endl; 
     while(current_node!=j) {
        //cerr<<endl<<"vertex: "<<coordinates[current_node]<<endl;
        //cerr<<"with basis "<<B<<endl;
        //we check first if there exists a suitable edge
        bool found=false;
        int best_node=current_node;
        Scalar best_improv(0);     
        if (find_way(current_node,Scalar(0),G,coordinates,j,best_node,best_improv)) {
          found=true;
          current_node=j;
        }
        else if (best_improv>0) {
           //cerr<<"existing edges"<<endl;
           current_node=best_node;
           B=one_basis[current_node];
           found=true;
        }
        if (!found) {
         bool vertex_unchanged=true;
         while (vertex_unchanged) {
          //reduced cost for the simplex (for maximizing objective into direction j)
          const Vector<Scalar> costs=lin_solve(T(A.minor(B,All)),-unit_vector<Scalar>(d,j));    
          Entire<Set<int> >::iterator iti=entire(B);
          //cerr<<"Costs: "<<costs<<endl;
          int index=0;
          while (costs[index]>=0) ++index,++iti;
          //cerr<<"index: "<<index<<"(i.e.: "<<(*iti)<<")"<<endl;
          //cerr<<"trying: "<<(A.minor(B,All))<<endl;
          const Vector<Scalar> dir=-lin_solve(A.minor(B,All),unit_vector<Scalar>(d,index));
          const Vector<Scalar> Ad=A*dir;
          const Vector<Scalar> Av=A*coordinates[current_node];
          //cerr<<"Direction: "<<dir<<endl<<"Ad: "<<Ad<<endl<<"Av: "<<Av<<endl;
          Scalar lambda;
          bool ray=true;
          int new_in=0;
          for (int k=0;k<n_facets;++k) {
             if (Ad[k]>0) 
	       if (ray) {
	         ray=false;
	         lambda=(b[k]-Av[k])/Ad[k];
	         new_in=k;
	       }
	     else {
	        const Scalar lambda_neu=(b[k]-Av[k])/Ad[k];
                //cerr<<"lambda "<<lambda_neu<<" for "<<k<<endl;
	        if (lambda_neu<lambda) {
	          new_in=k;
	          lambda=lambda_neu;
	        }
	     } 
          }
          if (lambda!=0) {
             vertex_unchanged=false;
             const Vector<Scalar> new_vertex=coordinates[current_node]+lambda*dir;
             typename hash_map<Vector<Scalar>,int >::const_iterator node=v_map.find(new_vertex);
             if (node==v_map.end()) {
                const int new_node=G.add_node();
                v_map[new_vertex]=new_node;
                G.edge(current_node,new_node);
                coordinates[new_node]=new_vertex;
                //cerr<<"new node "<<new_node<<endl<<new_vertex<<endl<<"new edge "<<current_node<<" to "<<new_node<<endl;
                B.erase(iti);
                B.insert(new_in);
                one_basis[new_node]=B;
                current_node=new_node;
             }
             else {
                const int new_node=node->second;
                G.edge(current_node,new_node);
                //cerr<<"new edge "<<current_node<<" to "<<new_node<<endl;
                current_node=new_node;
                //we could also use the still existing one_basis here
                B.erase(iti);
                B.insert(new_in);
             }
          }
          else {// only the basis has to be changed
             B.erase(iti);
             B.insert(new_in);
          }
         }
        }     
      }
      
  }

  for(Entire<Edges<Graph<> > >::const_iterator e=entire(edges(G)); !e.at_end(); ++e)
    edge_length[*e]=accumulate(attach_operation(coordinates[e.to_node()]-coordinates[e.from_node()],operations::abs_value()),operations::max());

  perl::Object g("WeightedLabeledGraph");
  g.take("ADJACENCY")<<G;
  g.take("N_NODES")<<G.nodes();
  g.take("EDGE_LENGTHS")<<edge_length;
  g.take("LABELED_NODES")<<elements;
  g.take("N_LABELED_NODES")<<d;
  return g;
}

FunctionTemplate4perl("find_realization(Matrix)"); 

} }
