/* Copyright (c) 2012 Sven Herrmann and Andreas Spillner

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "polymake/client.h"
#include "polymake/Map.h"
#include "polymake/Rational.h"
#include "polymake/Array.h"
#include "polymake/SparseMatrix.h"
#include "polymake/PowerSet.h"
#include "polymake/RandomGenerators.h"

namespace polymake { namespace metric {


typedef std::pair<Set<int>, Rational> WSplit;


SparseVector<Rational> split_metric(int n, const Set<int>& A)
{
   SparseVector<Rational> d(Integer::binom(n,2).to_int());
   const Set<int> A_compl=sequence(0,n)-A;
   int i=0;
   for (Subsets_of_k<const sequence&>::const_iterator s=entire(all_subsets_of_k(sequence(0,n),2)); !s.at_end(); ++s, ++i) {
      if (!(*s*A).empty() && !(*s*A_compl).empty()) d[i]=1;
   }
   
   return d;
}
      

SparseVector<Rational> split_system_metric(int n,const Array<WSplit>& splits)
{
  SparseVector<Rational> d(Integer::binom(n,2).to_int());
  for (int i=0; i<splits.size(); ++i) 
    d+=splits[i].second*split_metric(n,splits[i].first);
  return d;
}

SparseVector<Rational> split_system_metric2(int n,const Array<Set<int> >& splits)
{
  Array<WSplit> S(splits.size());
  for (int i=0; i<splits.size(); ++i)
    S[i]=WSplit(splits[i],1);
  return split_system_metric(n,S);
}

SparseVector<Rational> split_system_rand_metric(int n,const Array<Set<int> >& splits, int b,perl::OptionSet options)
{  
  RandomSeed seed(options["seed"]);
  UniformlyRandom<Integer> random_source(seed);
  Array<WSplit> S(splits.size());
  for (int i=0; i<splits.size(); ++i)
    S[i]=WSplit(splits[i],random_source.get()%b+1);
  return split_system_metric(n,S);
}




UserFunction4perl("# @category Producing from scratch"
		  "# Computes the split (pseudo) metric of a single split."
                  "# Output is a vector that can be used as DISTANCES property to obtain a proper metric object."
		  "# @param Int n cardinality of the space"
		  "# @param Set<Int> split (only one of the sets has to be given)"
		  "# @return SparseVector<Rational>",
		  &split_metric, "split_metric");

UserFunction4perl("# @category Producing from scratch"
		  "# Computes a metric from a weightend split system //splits//."
                  "# Each splits is given as a pair of one of the sets and the weight."
                  "# Output is a vector that can be used as DISTANCES property to obtain a proper metric object."
		  "# @param Int n cardinality of the space"
		  "# @param Array<Pair<Set<Int>, Rational> > splits"
		  "# @return SparseVector<Rational>",
		  &split_system_metric, "weighted_split_system_metric");
		  

UserFunction4perl("# @category Producing from scratch"
		  "# Computes a metric from a split system //splits// assigning each"
                  "# split a random integer weight between 1 and b."
                  "# Output is a vector that can be used as DISTANCES property to obtain a proper metric object."
		  "# @param Int n cardinality of the space"
		  "# @param Array<Set<Int> > splits (one of the set for each split)"
                  "# @param Int b"
		  "# @return SparseVector<Rational>",
		  &split_system_rand_metric, "split_system_rand_metric($$$;{ seed=>undef})");
		  


UserFunction4perl("# @category Producing from scratch"
		  "# Computes a metric from a split system //splits//."
                  "# Output is a vector that can be used as DISTANCES property to obtain a proper metric object."
		  "# @param Int n cardinality of the space"
		  "# @param Array<Set<int> > splits (one of the set for each split)"
		  "# @return SparseVector<Rational>",
		  &split_system_metric2, "split_system_metric");
		  
} }
