/* Copyright (c) 2012 Sven Herrmann and Andreas Spillner
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version: http://www.gnu.org/licenses/gpl.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
import java.util.*;
import java.math.*;

public class ConstraintFinder
{
//*******************************************************************************
//Finding the generating set Z
//*******************************************************************************

   public static TreeSet findGeneratingSet(TreeSet plist)
   {
      TreeSet genPairs = new TreeSet(new PairComparator());

      findVerticalPairs(plist,genPairs);
      findHorizontalPairs(plist,genPairs);
      findQuadrantPairs(plist,genPairs);

      return genPairs;
   }

   private static void findVerticalPairs(TreeSet plist,TreeSet genPairs)
   {
      TreeSet pv = new TreeSet(new PointLexComparatorX());

      pv.addAll(plist);

      Point cur = null;
      Point prev = null;
      Point lastbot1 = null;
      Point lastbot2 = null;

      Iterator iter = pv.iterator();

      while(iter.hasNext())
      {
         cur = (Point)iter.next();

         if((prev == null) || ((prev.x).compareTo(cur.x) == -1))
         {
            //moved one x-coordinate to the right
            if((lastbot1 != null) && (prev != null) && ((lastbot1.y).compareTo(prev.y) == 1))
            {
               genPairs.add(new Pair(lastbot1,prev));
            }

            if((prev != null) && ((prev.y).compareTo(cur.y) == -1))
            {
               genPairs.add(new Pair(prev,cur));
            }

            lastbot1 = lastbot2;
            lastbot2 = cur;
         }
         else
         {
            genPairs.add(new Pair(prev,cur));
         }

         prev = cur;
      }      
   }

   private static void findHorizontalPairs(TreeSet plist,TreeSet genPairs)
   {
      TreeSet ph = new TreeSet(new PointLexComparatorY());

      ph.addAll(plist);

      Point cur = null;
      Point prev = null;
      Point lastleft1 = null;
      Point lastleft2 = null;

      Iterator iter = ph.iterator();

      while(iter.hasNext())
      {
         cur = (Point)iter.next();

         if((prev == null) || ((prev.y).compareTo(cur.y) == -1))
         {
            //moved one y-coordinate to the top
            if((lastleft1 != null) && (prev != null) && ((lastleft1.x).compareTo(prev.x) == 1))
            {
               genPairs.add(new Pair(lastleft1,prev));
            }

            if((prev != null) && ((prev.x).compareTo(cur.x) == -1))
            {
               genPairs.add(new Pair(prev,cur));
            }

            lastleft1 = lastleft2;
            lastleft2 = cur;
         }
         else
         {
            genPairs.add(new Pair(prev,cur));
         }

         prev = cur;
      }      
   }

   private static void findQuadrantPairs(TreeSet plist,TreeSet genPairs)
   {
      TreeSet verHorPairs = new TreeSet(new PairComparator());
      verHorPairs.addAll(genPairs);

      TreeSet qlist = new TreeSet(new PointLexComparatorX());

      Iterator iter = plist.iterator();
  
      while(iter.hasNext())
      {
         Point p = (Point)iter.next();

         //first quadrant
         findPointsInQuadrant(plist,qlist,p,1,1);
         if(!quadrantIsIrrelevant(p,qlist,verHorPairs))
         {
            Point qmin = findMinimumPair(p,qlist,1,1);
            if(qmin != null) {genPairs.add(new Pair(p,qmin));}
         }

         //second quadrant
         findPointsInQuadrant(plist,qlist,p,1,-1);
         if(!quadrantIsIrrelevant(p,qlist,verHorPairs))
         {
            Point qmin = findMinimumPair(p,qlist,1,-1);
            if(qmin != null) {genPairs.add(new Pair(p,qmin));}
         }

         //third quadrant
         findPointsInQuadrant(plist,qlist,p,-1,-1);
         if(!quadrantIsIrrelevant(p,qlist,verHorPairs))
         {
            Point qmin = findMinimumPair(p,qlist,-1,-1);
            if(qmin != null) {genPairs.add(new Pair(p,qmin));}
         }

         //fourth quadrant
         findPointsInQuadrant(plist,qlist,p,-1,1);
         if(!quadrantIsIrrelevant(p,qlist,verHorPairs))
         {
            Point qmin = findMinimumPair(p,qlist,-1,1);
            if(qmin != null) {genPairs.add(new Pair(p,qmin));}
         }
      }
   }

   private static void findPointsInQuadrant(TreeSet plist,TreeSet qlist,Point p,int compx,int compy)
   {
      Iterator iter = plist.iterator();
      qlist.clear();

      while(iter.hasNext())
      {
         Point a = (Point)iter.next();
         if((((a.x).compareTo(p.x) == compx) || (a.x).equals(p.x)) && (((a.y).compareTo(p.y) == compy) || (a.y).equals(p.y)))
         {
            if(!((a.x).equals(p.x) && (a.y).equals(p.y))) {qlist.add(a);}
         }
      }
   }

   private static boolean quadrantIsIrrelevant(Point p,TreeSet qlist,TreeSet verHorPairs)
   {
      Iterator iter = qlist.iterator();
  
      while(iter.hasNext())
      {
         Point q = (Point)iter.next();
         if(verHorPairs.contains(new Pair(p,q))) 
         {
            return true;
         }
      }
      //System.out.println("Found a relevant quadrant: " + qlist.size());
      return false;
   }

   private static Point findMinimumPair(Point p,TreeSet qlist,int compx,int compy)
   {
      Iterator iter = qlist.iterator();
  
      Point qmin = null;

      while(iter.hasNext())
      {
         Point q = (Point)iter.next();

         if((qmin == null) || ((qmin.x).compareTo(q.x) == compx) || ((qmin.y).compareTo(q.y) == compy))
         {
            qmin = q;
         }
      }
      return qmin;
   }

//****************************************************************************************
//setting up the grid graph
//****************************************************************************************

   public static LinkedList setUpGridGraph(TreeSet plist, TreeSet genPairs)
   {
      LinkedList edges = new LinkedList();

      TreeSet xcoord = collectXCoordinates(plist);
      TreeSet ycoord = collectYCoordinates(plist);

      Iterator xiter = null;
      Iterator yiter = null;

      //create vertical edges
      xiter = xcoord.iterator();
      while(xiter.hasNext())
      {
         BigInteger x = (BigInteger)xiter.next();
         yiter = ycoord.iterator();
         Point prev = null;
         Point cur = null;
         while(yiter.hasNext())
         {
            BigInteger y = (BigInteger)yiter.next();
            cur = new Point(x,y);
            if(prev != null) {edges.add(new Edge(prev,cur));}
            prev = cur;
         }
      }

      //create horizontal edges
      yiter = ycoord.iterator();
      while(yiter.hasNext())
      {
         BigInteger y = (BigInteger)yiter.next();
         xiter = xcoord.iterator();
         Point prev = null;
         Point cur = null;
         while(xiter.hasNext())
         {
            BigInteger x = (BigInteger)xiter.next();
            cur = new Point(x,y);
            if(prev != null) {edges.add(new Edge(prev,cur));}
            prev = cur;
         }
      }

      attachPairsToEdges(edges,genPairs);
      computeOffsets(edges);

      return edges;
   }

   private static TreeSet collectXCoordinates(TreeSet plist)
   {
      Iterator iter = plist.iterator();
      TreeSet xcoord = new TreeSet();

      while(iter.hasNext())
      {
         Point p = (Point)iter.next();
         xcoord.add(p.x);
      }
     
      return xcoord;
   }

   private static TreeSet collectYCoordinates(TreeSet plist)
   {
      Iterator iter = plist.iterator();
      TreeSet ycoord = new TreeSet();

      while(iter.hasNext())
      {
         Point p = (Point)iter.next();
         ycoord.add(p.y);
      }
     
      return ycoord;
   }

   private static void attachPairsToEdges(LinkedList edges,TreeSet genPairs)
   {
      Iterator eiter = edges.iterator();

      while(eiter.hasNext())
      {
         Edge e = (Edge)eiter.next();

         Iterator piter = genPairs.iterator();

         while(piter.hasNext())
         {
            Pair a = (Pair)piter.next();
            if(isInBoundingBox(e.u,a) && isInBoundingBox(e.v,a))
            {
               if((e.u.y).equals(e.v.y))
               {
                  //horizontal edge!
                  e.uvpairs.add(a);
               }
               else if((a.p.y).compareTo(a.q.y) == -1)
               {
                  e.uvpairs.add(a);
               }
               else
               {
                  e.vupairs.add(a);
               }
            }
         }
      } 
   }

   private static boolean isInBoundingBox(Point p,Pair a)
   {
      if(((a.p.x).compareTo(p.x) <= 0) && ((a.q.x).compareTo(p.x) >= 0))
      {
         if((((a.p.y).compareTo(p.y) <= 0) && ((a.q.y).compareTo(p.y) >= 0)) || (((a.q.y).compareTo(p.y) <= 0) && ((a.p.y).compareTo(p.y) >= 0)))
         {
            //System.out.println("("+p.x.toString()+","+p.y.toString()+") is in ((" + a.p.x.toString() + "," + a.p.y.toString() + "),(" + a.q.x.toString() + "," + a.q.y.toString() + "))");
            return true;
         }
      }
      //System.out.println("("+p.x.toString()+","+p.y.toString()+") is not in ((" + a.p.x.toString() + "," + a.p.y.toString() + "),(" + a.q.x.toString() + "," + a.q.y.toString() + "))");
      return false;
   }

   private static void computeOffsets(LinkedList edges)
   {
      Iterator iter = edges.iterator();

      int curidx = 1;      

      //find out which edges in the grid are actually used
      while(iter.hasNext())
      {
         Edge e = (Edge)iter.next();
         if((e.uvpairs.size() + e.vupairs.size()) > 0)
         {
            e.idx = curidx;
            curidx++;
         }
         else
         {
            e.idx = -1;
         }
      }

      //compute how many constraints come before this edge
      iter = edges.iterator();
      while(iter.hasNext())
      {
         Edge e = (Edge)iter.next();
         e.offset = curidx;
         curidx = curidx + e.uvpairs.size() + e.vupairs.size();
      }
   }

   public static LinkedList getGridVertices(TreeSet plist,LinkedList edges)
   {
      LinkedList vertices = new LinkedList();

      TreeSet xcoord = collectXCoordinates(plist);
      TreeSet ycoord = collectYCoordinates(plist);

      Iterator xiter = null;
      Iterator yiter = null;

      xiter = xcoord.iterator();
      while(xiter.hasNext())
      {
         BigInteger x = (BigInteger)xiter.next();
         yiter = ycoord.iterator();
         Point cur = null;
         while(yiter.hasNext())
         {
            BigInteger y = (BigInteger)yiter.next();
            cur = new Point(x,y);
            Vertex w = new Vertex(cur);
            findIncidentEdges(w,edges);
            vertices.add(w);
         }
      }
      return vertices;
     
   }

   private static void findIncidentEdges(Vertex w,LinkedList edges)
   {
      Iterator iter = edges.iterator();

      while(iter.hasNext())
      {
         Edge e = (Edge)iter.next();

         if(((e.u.x).equals(w.p.x) && (e.u.y).equals(w.p.y)) || ((e.v.x).equals(w.p.x) && (e.v.y).equals(w.p.y)))
         {
            w.elist.add(e);
         }
      }
   }
}
